import geopandas as gpd
import numpy as np
import math

from ..processing import read_raster
from ..processing import normalised_difference
from ..processing import mask_raster

def water_quality(image_name,aoi_mask,storage):
    # read_raster image
    ds, raster = read_raster(image_name, storage, bands=[3,4,5,8])
    date_and_tif = image_name.split("_")[1]
    date = date_and_tif.split(".")[0]
    # calculate ndwi
    ndwi = normalised_difference(raster, [1,4])

    # save_raster ndwi
    
    raster_name_ndwi = f"ndwi_{date}.tif"
    storage.create(ndwi, raster_name_ndwi, ds=ds)

    # read_geojson aoi_mask
    aoi_mask_gdf = gpd.GeoDataFrame.from_features(aoi_mask["features"], crs=4326)

    # mask_raster ndwi with aoi_mask
    ndwi_masked, out_transform = mask_raster(raster_name_ndwi, aoi_mask_gdf, storage)

    # save_raster ndwi_masked
    raster_name_ndwi_masked = f"ndwi_masked_{date}.tif"
    storage.create(ndwi_masked, raster_name_ndwi_masked, ds=ds)

    # apply thershold to calculate water-mask
    threshold = -0.1
    water_mask = ndwi >= threshold

    # save_raster water_mask
    raster_name_water_mask = f"water_mask_{date}.tif"
    storage.create(water_mask, raster_name_water_mask, ds=ds)

    # mask_raster water_mask with aoi_mask
    water_mask_masked, out_transform = mask_raster(raster_name_water_mask, aoi_mask_gdf, storage)

    # save_raster water_mask_masked
    raster_name_water_mask_masked = f"water_mask_masked_{date}.tif"
    storage.create(water_mask_masked, raster_name_water_mask_masked, ds=ds)


    # calculate ndti
    ndti = normalised_difference(raster, [2,1])

    # save_raster ndti
    raster_name_ndti = f"ndti_{date}.tif"
    storage.create(ndti, raster_name_ndti, ds=ds)

    # mask_raster ndti with aoi_mask
    ndti_masked, out_transform = mask_raster(raster_name_ndti, aoi_mask_gdf, storage)

    # save_raster ndti_masked
    raster_name_ndti_masked = f"ndti_masked_{date}.tif"
    storage.create(ndti_masked, raster_name_ndti_masked, ds=ds)

    # calculate ndci
    ndci = normalised_difference(raster, [3,2])

    # save_raster ndci
    raster_name_ndci = f"ndci_{date}.tif"
    storage.create(ndci, raster_name_ndci, ds=ds)

    # mask_raster ndci with aoi_mask
    ndci_masked, out_transform = mask_raster(raster_name_ndci, aoi_mask_gdf, storage)

    # save_raster ndci_masked
    raster_name_ndci_masked = f"ndci_masked_{date}.tif"
    storage.create(ndci_masked, raster_name_ndci_masked, ds=ds)

    # calculate doc
    bands = [1,2]
    bands = np.array(bands) - 1
    # Separate the bands
    band1 = raster[bands[0], :, :]
    band2 = raster[bands[1], :, :]
    # convert the bands to floats
    band1 = band1.astype(float)
    band2 = band2.astype(float)
    doc = 432 * pow(math.e, -2.24 * band1/band2)

    # save_raster doc
    raster_name_doc = f"doc_{date}.tif"
    storage.create(doc, raster_name_doc, ds=ds)

    # mask_raster doc with aoi_mask
    doc_masked, out_transform = mask_raster(raster_name_doc, aoi_mask_gdf, storage)

    # save_raster doc_masked
    raster_name_doc_masked = f"doc_masked_{date}.tif"
    storage.create(doc_masked, raster_name_doc_masked, ds=ds)


import pandas as pd
import json
import io
from minio import Minio
import fnmatch

from .BaseStorage import BaseStorage
from .decorators import with_rio, with_geopandas


class S3Storage(BaseStorage):
    def __init__(self, url, access, secret, bucket, region=None):
        super().__init__()
        self.url = url
        self.access = access
        self.secret = secret
        self.region = region
        self.bucket = bucket

        if self.access and self.secret:
            # Create a client
            self.client = Minio(
                endpoint=self.url,
                access_key=self.access,
                secret_key=self.secret,
                secure=True if self.region else False,
                region=self.region,
            )  # because no certificate is used in the containerised version of minio
            if not self.client.bucket_exists(self.bucket):
                # Make a bucket with the credentials and the bucket_name given
                self.client.make_bucket(self.bucket)
                print(f"Bucket '{self.bucket}' created")
            else:
                print(f"'{self.bucket}' bucket ready to use")
        else:
            # TODO: create bucket in our minio server (we will need our credentials for that, do it with API request?
            print("Missing credentials")
            # Habría que preguntar si se quiere crear el bucket en nuestro cloud o decirles que introduzcan sus creds

    def list(self, pattern="*"):
        return fnmatch.filter(
            [
                obj.object_name
                for obj in self.client.list_objects(self.bucket, recursive=True)
            ],
            pattern,
        )

    def create_from_path(self, data, name):
        if data.endswith(".json"):
            content_type = "application/json"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return name
        elif data.endswith(".tiff") or data.endswith(".tif"):
            content_type = "image/tiff"
            # with rasterio.open(data) as src:
            #     self.client.put_object(self.bucket, dst_path, src, length=src.read().all(), content_type=content_type)
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return name
        elif data.endswith(".geojson"):
            content_type = "application/geojson"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return name
        else:
            print(data)
            # self.client.put_object(self.bucket, prefix + name, data, length, content_type)

    def create_from_dict(self, data, name):
        if name.endswith(".json"):
            content_type = "application/json"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".geojson"):
            content_type = "application/geojson"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dict type extension")

    def create_from_string(self, data, name):
        content_type = "text/plain"
        prefix = "txt/"
        content = data.encode("utf8")
        self.client.put_object(
            self.bucket,
            name,
            io.BytesIO(content),
            -1,
            part_size=50 * 1024 * 1024,
            content_type=content_type,
        )
        return name

    def create_from_dataframe(self, data, name):
        if name.endswith(".csv"):
            content_type = "text/csv"
            content = data.to_csv().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".json"):
            content_type = "application/json"
            content = data.to_json().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dataframe type extension")

    def read_from_json(self, name):
        try:
            response = self.client.get_object(self.bucket, name)
            data = json.load(response)
        finally:
            response.close()
            response.release_conn()
        df = pd.DataFrame.from_dict(data)
        # Convert the index to integers (the timestamps) and then to datetime
        df.index = pd.to_datetime(df.index.astype(int), unit="ms")
        # Now your index is a datetime object and you can use strftime
        df.index = df.index.strftime("%Y-%m-%d")
        return df

    @with_geopandas
    def read_from_geojson(self, gpd, name):
        try:
            response = self.client.get_object(self.bucket, name)
            data = json.load(response)
        finally:
            response.close()
            response.release_conn()
        return gpd.GeoDataFrame.from_features(data)

    @with_rio
    def read_from_rasterio(self, rio, name):
        try:
            response = self.client.get_object(self.bucket, name)
            data = response
        finally:
            response.close()
            response.release_conn()
        return rio.open(data)



from ..repos import APIRepo
import os


def deploy_template(user, template_name, typer):
    repo = APIRepo()
    data, error = repo.deploy_template(user, template_name)
    if data:
        typer.echo(
            # f"Check status at {os.getenv('SPAI_UI_URL', 'https://spai.earthpulse.ai')}/dashboard/{data}"
            f"Check status at {os.getenv('SPAI_UI_URL', 'https://spai-new.vercel.app')}/dashboard/{data['project_id']}"
        )
        return
    raise Exception(f"Error: {error}")



import pandas as pd
import json
import io
from minio import Minio
import fnmatch
from io import BytesIO
import numpy as np

from .BaseStorage import BaseStorage
from .decorators import with_rio, with_geopandas


# Ver funciones en CloudStorage si las de aquí no acaban de funcionar...


class S3Storage(BaseStorage):
    def __init__(self, url, access, secret, bucket, region=None):
        super().__init__()
        self.url = url
        self.access = access
        self.secret = secret
        self.region = region
        self.bucket = bucket

        if self.access and self.secret:
            # Create a client
            self.client = Minio(
                endpoint=self.url,
                access_key=self.access,
                secret_key=self.secret,
                secure=True if self.region else False,
                region=self.region,
            )  # because no certificate is used in the containerised version of minio
            if not self.client.bucket_exists(self.bucket):
                # Make a bucket with the credentials and the bucket_name given
                self.client.make_bucket(self.bucket)
                print(f"Bucket '{self.bucket}' created")
            else:
                print(f"'{self.bucket}' bucket ready to use")
        else:
            # TODO: create bucket in our minio server (we will need our credentials for that, do it with API request?
            print("Missing credentials")
            # Habría que preguntar si se quiere crear el bucket en nuestro cloud o decirles que introduzcan sus creds

    def get_path(self, name):
        return self.get_url(name)

    def get_url(self, name):
        return self.client.presigned_get_object(self.bucket, name)

    def list(self, pattern="*"):
        return fnmatch.filter(
            [
                obj.object_name
                for obj in self.client.list_objects(self.bucket, recursive=True)
            ],
            pattern,
        )

    def create_from_path(self, data, name):
        if data.endswith(".json"):
            content_type = "application/json"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        if data.endswith(".tiff") or data.endswith(".tif"):
            content_type = "image/tiff"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        if data.endswith(".geojson"):
            content_type = "application/geojson"
            self.client.fput_object(self.bucket, name, data, content_type=content_type)
            return self.get_url(name)
        self.client.fput_object(self.bucket, name, data)
        return self.get_url(name)

    def create_from_image(self, data, name):
        image_bytes = BytesIO()
        # get format from file extension
        format = name.split(".")[-1]
        data.save(image_bytes, format=format)
        image_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, image_bytes, len(image_bytes.getvalue())
        )
        return self.get_url(name)

    @with_rio
    def create_from_rasterio(self, rio, x, name, ds, window=None):
        transform = ds.transform if window is None else ds.window_transform(window)
        with rio.io.MemoryFile() as memfile:
            with memfile.open(
                driver="GTiff",
                count=1 if x.ndim < 3 else x.shape[0],
                height=x.shape[0] if x.ndim < 3 else x.shape[1],
                width=x.shape[1] if x.ndim < 3 else x.shape[2],
                dtype=np.uint8 if x.dtype == "bool" else x.dtype,
                crs=ds.crs,
                transform=transform,
                # nbits=1 if x.dtype == 'bool' else
            ) as dest_ds:
                bands = 1 if x.ndim < 3 else [i + 1 for i in range(x.shape[0])]
                dest_ds.write(x, bands)
            self.client.put_object(
                self.bucket, name, memfile, length=-1, part_size=10 * 1024 * 1024
            )
        return self.get_url(name)

    def create_from_array(self, data, name):
        array_bytes = BytesIO()
        np.save(array_bytes, data)
        array_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, array_bytes, len(array_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_csv(self, data, name):
        csv_bytes = BytesIO()
        data.to_csv(csv_bytes)
        csv_bytes.seek(0)
        self.client.put_object(self.bucket, name, csv_bytes, len(csv_bytes.getvalue()))
        return self.get_url(name)

    def create_from_json(self, data, name):
        json_bytes = BytesIO()
        data.to_json(json_bytes)
        json_bytes.seek(0)
        self.client.put_object(
            self.bucket, name, json_bytes, len(json_bytes.getvalue())
        )
        return self.get_url(name)

    def create_from_dict(self, data, name):
        if name.endswith(".json"):
            content_type = "application/json"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".geojson"):
            content_type = "application/geojson"
            content = json.dumps(data, ensure_ascii=False).encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dict type extension")

    def create_from_string(self, data, name):
        content_type = "text/plain"
        prefix = "txt/"
        content = data.encode("utf8")
        self.client.put_object(
            self.bucket,
            name,
            io.BytesIO(content),
            -1,
            part_size=50 * 1024 * 1024,
            content_type=content_type,
        )
        return name

    def create_from_dataframe(self, data, name):
        if name.endswith(".csv"):
            content_type = "text/csv"
            content = data.to_csv().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        elif name.endswith(".json"):
            content_type = "application/json"
            content = data.to_json().encode("utf8")
            self.client.put_object(
                self.bucket,
                name,
                io.BytesIO(content),
                -1,
                part_size=50 * 1024 * 1024,
                content_type=content_type,
            )
            return name
        else:
            raise TypeError("Not a valid dataframe type extension")

    def read_object(self, name):
        return BytesIO(self.client.get_object(self.bucket, name).read())

    def read_from_json(self, name):
        return pd.read_json(self.read_object(name))

    def read_from_array(self, name):
        return np.load(self.read_object(name))

    @with_geopandas
    def read_from_geojson(self, gpd, name):
        try:
            response = self.client.get_object(self.bucket, name)
            data = json.load(response)
        finally:
            response.close()
            response.release_conn()
        return gpd.GeoDataFrame.from_features(data)

    @with_rio
    def read_from_rasterio(self, rio, name):
        return rio.open(self.read_object(name))

    def read_from_csv(self, name):
        return pd.read_csv(self.read_object(name), index_col=0)



from pydantic import BaseModel
from typing import Union, List, Any, Dict
from pathlib import Path
from typing import Optional
import json

from .StorageConfig import LocalStorageConfig, S3StorageConfig


class ScriptConfig(BaseModel):
    name: str
    run_on_start: bool = True
    command: Optional[str] = None
    run_every: Optional[int] = None  # seconds (in cloud minutes)
    storage: Optional[str] = None  # folder to bind in cloud
    type: str = "script"


class NotebookConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    storage: Union[str, None] = None  # folder to bind in cloud
    port: int = 8888
    host: str = "0.0.0.0"
    type: str = "notebook"


class APIConfig(BaseModel):
    name: str
    command: Union[str, None] = None
    port: int = 8000
    host: str = "0.0.0.0"
    storage: Union[str, None] = None  # folder to bind in cloud
    type: str = "api"


class UIConfig(BaseModel):
    name: str
    command: str  # steamlit, javascript, ...
    port: int = 3000
    host: str = "0.0.0.0"
    env: dict = {}  # can accept the name of another service as a url placeholder
    type: str = "ui"


class Config(BaseModel):
    dir: Path
    project: str
    scripts: List[ScriptConfig] = []
    notebooks: List[NotebookConfig] = []
    apis: List[APIConfig] = []
    uis: List[UIConfig] = []
    storage: List[Union[LocalStorageConfig, S3StorageConfig]] = []

    # iterator for all the services
    def __iter__(self):
        # if self.storage:
        #     for storage in self.storage:
        #         yield storage
        if self.scripts:
            for script in self.scripts:
                yield script
        if self.notebooks:
            for notebook in self.notebooks:
                yield notebook
        if self.apis:
            for api in self.apis:
                yield api
        if self.uis:
            for ui in self.uis:
                yield ui
        if self.storage:
            for storage in self.storage:
                yield storage

    def type2folder(self, type):
        return type + "s"


# class Config:

#     def __init__(self, dir, config_filename="spai.config.json"):
#         self.dir = Path(dir)
#         self.config_path = self.dir / config_filename
#         self.config = self.load_config(self.config_path)
#         self.project = self.config.get('project', None)
#         self.scripts = [ScriptConfig(**script) for script in self.config.get('scripts', [])]
#         self.notebooks = [NotebookConfig(**notebook) for notebook in self.config.get('notebooks', [])]
#         self.apis = [APIConfig(**api) for api in self.config.get('apis', [])]
#         self.uis = [UIConfig(**ui) for ui in self.config.get('uis', [])]
#         self.storage = [LocalStorageConfig(**storage) if storage['type'] == 'local' else S3StorageConfig(**storage) for storage in self.config.get('storage', [])]

#     def load_config(self, config_path):
#         if config_path.exists():
#             with open(config_path, "r") as f:
#                 try:
#                     return json.load(f)
#                 except json.JSONDecodeError as e:
#                     print(f"Error loading JSON from {config_path}: {e}")
#         else:
#             print(f"Config file not found: {config_path}")
#         return {}
